/*
 * dbook.c --
 *
 *	Definition for packer wrappers to datebooks.
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: dbook.c,v 1.1 1999/03/27 13:26:32 aku Exp $
 */

#include "wrap.h"
#include "dates.h"
#include "category.h"
#include "internal.h"

#include "pi-datebook.h"

typedef struct Appointment         Appointment;
typedef struct AppointmentAppInfo  AppointmentAppInfo;


/*
 * Declarations of internal custom converter procedures.
 */

static Tcl_Obj*
ExcToObj _ANSI_ARGS_ ((Tcl_Interp* interp,
		       void*       itemPtr,
		       ClientData  clientData));

static int
ExcFromObj _ANSI_ARGS_ ((Tcl_Interp* interp,
			 Tcl_Obj*    o,
			 void*       itemPtr,
			 ClientData  clientData));

static Tcl_Obj*
DaysToObj _ANSI_ARGS_ ((Tcl_Interp* interp,
			void*       itemPtr,
			ClientData  clientData));

static int
DaysFromObj _ANSI_ARGS_ ((Tcl_Interp* interp,
			  Tcl_Obj*    o,
			  void*       itemPtr,
			  ClientData  clientData));

/*
 * Definitions for conversion of todo records.
 */

static TclUtil_EnumItem advValues [] = {
  {"min",  advMinutes},
  {"hour", advHours},
  {"day",  advDays},
  {NULL,  -1},
};

static TclUtil_Enum advEnum = {
  "advance unit", NULL, advValues
};


static TclUtil_EnumItem reptValues [] = {
  {"none",          repeatNone},
  {"daily",         repeatDaily},
  {"weekly",        repeatWeekly},
  {"monthlyByDay",  repeatMonthlyByDay},
  {"monthlyByDate", repeatMonthlyByDate},
  {"yearly",        repeatYearly},
  {NULL,           -1},
};

static TclUtil_Enum reptEnum = {
  "repeat type", NULL, reptValues
};


static TclUtil_EnumItem repWsValues [] = {
  {"sunday",  0},
  {"monday",  1},
  {NULL,     -1},
};

static TclUtil_Enum repWsEnum = {
  "repeat weekstart", NULL, repWsValues
};


/* The master/sub definition are a bit hackish to break
 * the main structure into simpler units. The sub-
 * structures are always anchored on the first field,
 * with offset 0, hence they start at the beginning
 * of the whole structure.
 */

static TclUtil_StructItem advItems [] = {
  ITEM        ("value", advance,      Appointment, TUSI_Int),
  ITEM_ENUM   ("unit",  advanceUnits, Appointment, &advEnum),
  ITEM_END
};

static TclUtil_Struct advStruct = {
  "advance", 0, NULL, advItems
};

static TclUtil_StructItem repItems [] = {
  ITEM_ENUM   ("type",      repeatType,      Appointment, &reptEnum),
  ITEM        ("freq",      repeatFrequency, Appointment, TUSI_Int),
  ITEM_ENUM   ("weekstart", repeatWeekstart, Appointment, &repWsEnum),
  ITEM        ("forever",   repeatForever,   Appointment, TUSI_Boolean),
  ITEM_CUSTOM ("end",       repeatEnd,       Appointment, PiTclDateToObj, PiTclDateFromObj, NULL),
  ITEM        ("day",       repeatDay,       Appointment, TUSI_Int),
  ITEM_CUSTOM ("exceptions", event,          Appointment, ExcToObj,  ExcFromObj,  NULL),
  ITEM_CUSTOM ("days",       event,          Appointment, DaysToObj, DaysFromObj, NULL),
  ITEM_END
};

static TclUtil_Struct repStruct = {
  "repeat", 0, NULL, repItems
};


static TclUtil_StructItem recordItems [] = {
  ITEM_CUSTOM ("begin",       begin,       Appointment, PiTclDateToObj, PiTclDateFromObj, NULL),
  ITEM_CUSTOM ("end",         end,         Appointment, PiTclDateToObj, PiTclDateFromObj, NULL),
  ITEM        ("description", description, Appointment, TUSI_String),
  ITEM        ("note",        note,        Appointment, TUSI_String),
  ITEM        ("event",       event,       Appointment, TUSI_Boolean),
  ITEM        ("alarm",       alarm,       Appointment, TUSI_Boolean),
  ITEM_STRUCT ("advance",     event,       Appointment, &advStruct),
  ITEM_STRUCT ("repeat",      event,       Appointment, &repStruct),
  ITEM_END
};

static TclUtil_Struct recordStruct = {
  "Appointment", 0, NULL, recordItems
};

static PiTcl_CvtData recordConverter = {
  &recordStruct,                         /* Definition of Appointment struc-
					  * ture, for conversion to tcl list */
  sizeof (Appointment),                  /* Size of the Appointment structure
					  */
  (PiTcl_FreeStruct*) free_Appointment,  /* Proc to free additionally allocated
					  * space */
  (PiTcl_CvtStruct*) pack_Appointment,   /* Serialize Appointment into byte-
					  * stream */
  (PiTcl_CvtStruct*) unpack_Appointment, /* Unserialize bytestream into an
					  * appointment */
  4096                                   /* Initial size of buffer to use for
					  * serialization */
};


/*
 * Definitions for conversion of Appointment application information.
 */

static TclUtil_StructItem appinfoItems [] = {
  ITEM_STRUCT ("category",    category,    AppointmentAppInfo, &catStruct),
  ITEM_ENUM   ("startOfWeek", startOfWeek, AppointmentAppInfo, &repWsEnum),
  ITEM_END
};

static TclUtil_Struct appinfoStruct = {
  "appointment application information", 0, NULL, appinfoItems
};

static PiTcl_CvtData appinfoConverter = {
  &appinfoStruct,                               /* Definition of appointment
						 * appinfo structure, for con-
						 * version to and from a tcl
						 * list */
  sizeof (AppointmentAppInfo),                  /* Size of the appointment
						 * appinfo structure */
  NULL,                                         /* *No* additionally allocated
						 * space */
  (PiTcl_CvtStruct*) pack_AppointmentAppInfo,   /* Serialize Appointment app-
						 * info into a bytestream */
  (PiTcl_CvtStruct*) unpack_AppointmentAppInfo, /* Unserialize bytestream into
						 * an appointment appinfo
						 * block */
  sizeof (AppointmentAppInfo)+1                 /* Initial size of buffer to
						 * use for serialization */
};

/*
 * Appointment databases do not have a sort information block.
 */

/*
 * Definition of the whole Appointment converter.
 */

static PiTcl_PilotCvt converter = {
  &recordConverter,
  NULL,
  &appinfoConverter
};

/*
 *------------------------------------------------------*
 *
 *	PiTclInitAppointment --
 *
 *	------------------------------------------------*
 *	This procedure registers the commands wrapped
 *	around the (un)pack functionality for Appointments.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		See the user documentation.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclInitAppointment (interp)
Tcl_Interp* interp;
{
  Tcl_CreateObjCommand (interp, "piAppointment",
			&PiTclPilotCvtCmd,
			(ClientData) &converter,
			NULL);

  return TCL_OK;
}



static Tcl_Obj*
ExcToObj (interp, itemPtr, clientData)
Tcl_Interp* interp;
void*       itemPtr;
ClientData  clientData;
{
  Appointment* a    = (Appointment*) itemPtr;
  Tcl_Obj*     list = Tcl_NewListObj (0, NULL);
  Tcl_Obj*     o;
  int          i, res;

  if (list == (Tcl_Obj*) NULL) {
    return (Tcl_Obj*) NULL;
  }

  for (i=0; i < a->exceptions; i++) {

      o = PiTclDateToObj (interp, &a->exception [i], NULL);

      if (o == (Tcl_Obj*) NULL) {
	Tcl_DecrRefCount (list);
	return (Tcl_Obj*) NULL;
      }

      res = Tcl_ListObjAppendElement (interp, list, o);

      if (res != TCL_OK) {
	Tcl_DecrRefCount (list);
	return (Tcl_Obj*) NULL;
      }

      /* check next */
  }
 
  return list;
}


static int
ExcFromObj (interp, o, itemPtr, clientData)
Tcl_Interp* interp;
Tcl_Obj*    o;
void*       itemPtr;
ClientData  clientData;
{
  Appointment* a = (Appointment*) itemPtr;
  Tcl_Obj**    lobjs;
  int          length, i, res;

  res = Tcl_ListObjGetElements (interp, o, &length, &lobjs);

  if (res != TCL_OK) {
    return TCL_ERROR;
  }

  a->exceptions = length;

  if (length == 0) {
    a->exception = NULL;
  } else {
    a->exception  = (struct tm*) Tcl_Alloc (length * sizeof (struct tm));

    for (i=0; i < length; i++) {
      res = PiTclDateFromObj (interp, lobjs [i], &a->exception [i], NULL);

      if (res != TCL_OK) {
	return TCL_ERROR;
      }
    }
  }

  return TCL_OK;
}



static Tcl_Obj*
DaysToObj (interp, itemPtr, clientData)
Tcl_Interp* interp;
void*       itemPtr;
ClientData  clientData;
{
  Appointment* a    = (Appointment*) itemPtr;
  char         buf [7];
  int          i;

  for (i=0; i < 7; i++) {
    buf [i] = (a->repeatDays [i] ? '1' : '0');
  }

  return Tcl_NewStringObj (buf, 7);
}



static int
DaysFromObj (interp, o, itemPtr, clientData)
Tcl_Interp* interp;
Tcl_Obj*    o;
void*       itemPtr;
ClientData  clientData;
{
  Appointment* a = (Appointment*) itemPtr;
  char*        buf;
  int          length, i;

  for (i=0; i < 7; i++) {
    a->repeatDays [i] = 0;
  }

  buf = Tcl_GetStringFromObj (o, &length);

  if (length > 7) {
    length = 7;
  }

  for (i=0; i < length; i++) {
    a->repeatDays [i] = (buf [i] == '0' ? 0 : 1);
  }

  return TCL_OK;
}
