#ifndef PITCL_INT_H
#define PITCL_INT_H
/* -*- c -*-
 *
 * piTclInt.h --
 *
 *	Global declarations of piTcl, internal to the package
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: internal.h,v 1.2 1999/06/28 21:20:35 aku Exp $
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <tcl.h>
#include <errno.h>

#include "pi-config.h" /* Access pilot-link configuration */
#include "pi-socket.h" /* Access the pilot-link socket layer */
#include "pi-dlp.h"    /* Access the pilot-link dlp layer */

#include "tclUtil.h"   /* Convenience functionality */

typedef struct CardInfo    CardInfo;
typedef struct NetSyncInfo NetSyncInfo;
typedef struct SysInfo     SysInfo;
typedef struct PilotUser   PilotUser;
typedef struct DBInfo      DBInfo;

#include "macros.h" /* Internal helper macro definitions */
#include "pack.h"
#include "map.h"

/* -- using the include above instead, assume configured pilot-link!
#if TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   if HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif
*/

#if 0
/* enable use of procedure internal to tcl
 * DISABLEED, currently not used, may conflict with
 * declaration in 8.0.4 pluspatch, ...
 */
EXTERN void
panic _ANSI_ARGS_ ((char* format, ...));
#endif


/*
 * Definitions to enable the generation of a DLL under Windows.
 * Taken from 'ftp://ftp.sunlabs.com/pub/tcl/example.zip(example.c)'
 */

#if defined(__WIN32__)
#   define WIN32_LEAN_AND_MEAN
#   include <windows.h>
#   undef WIN32_LEAN_AND_MEAN

/*
 * VC++ has an alternate entry point called DllMain, so we need to rename
 * our entry point.
 */

#ifdef TCL_STORAGE_CLASS
# undef TCL_STORAGE_CLASS
#endif
#ifdef BUILD_pitcl
# define TCL_STORAGE_CLASS DLLEXPORT
#else
# define TCL_STORAGE_CLASS DLLIMPORT
#endif

#   if defined(_MSC_VER)
#	define EXPORT(a,b) TCL_STORAGE_CLASS a b
#	define DllEntryPoint DllMain
#   else
#	if defined(__BORLANDC__)
#	    define EXPORT(a,b) a _export b
#	else
#	    define EXPORT(a,b) a b
#	endif
#   endif
#else
#   define EXPORT(a,b) a b
#endif




/* Global structure. Records all devices currently used for communication with
 * pilots, to prevent multiple use. Given as clientData-pointer to all object
 * generator commands.
 */

typedef struct PiTclInfo {
  int           refCount; /* The structure can be freed if this drops to or
			   * below 0 */
  Tcl_HashTable objects;  /* Refers to all created objects, is indexed with
			   * the name of the used devices */

  PiTclPackers  pack;     /* Refers to all registered packer structures */
  PiTclMap      map;      /* Mapping from databases to converters */

} PiTclInfo;


/* Common structure to use as prefix in the definitions of object instance
 * structures. Required to allow the cleanup of the global structure to
 * annihilate the stored backreference pointer. And that one is required to
 * allow object instances to remove themselves from the global locktable.
 */

typedef struct PiTclObjInfo {
  PiTclInfo* info; /* Backreference to the global datastructures */
} PiTclObjInfo;




/* Declare procedures internal to piTcl, but required
 * in different places.
 */

/* Check for usage of given device, to prevent simultaneous access to one
 * pilot by more than one object.
 */

EXTERN int
PiTclDeviceIsInUse _ANSI_ARGS_ ((PiTclInfo* info, char* deviceName));

EXTERN void
PiTclDeleteCommand _ANSI_ARGS_ ((ClientData cd));


/* Object generator procedure, creates DLP objects.
 * The clientData refers to the global 'PiTclInfo'
 * structure.
 */

EXTERN int
PiTclDlpCmd _ANSI_ARGS_ ((ClientData     cd,
			  Tcl_Interp*    interp,
			  int            objc,
			  Tcl_Obj* CONST objv []));


/*
 * This structure describes the per-instance state of a DLP object
 */

typedef struct PiTclDlpInstance {

  PiTclObjInfo base; /* Information required by the global structures */
  char*        dev;  /* Name of the device we are connected to */
  int          sd;   /* Socket handle returned during socket creation */
  int  acceptDone;   /* Boolean flag. Initially 0. Set to 1 after 'pi_accept'
		      * returned successfully. */
  int  inSession;    /* Boolean flag. Initially 0. Set/reset after the
		      * start/shutdown of a conduit session. */
  Tcl_Command cmd;   /* Reference to our object command, for easy deletion */
  Tcl_Interp* interp; /* Reference to the interpreter holding the command */

  /* Cache containing card information */

  int         numCards;
  CardInfo** card;

  /* Table of database objects, indexed by database name
   * Contains all! databases known to the pilot we are connected to.
   * A flag is used to mark open databases.
   */

  int           dbInitialized; /* Defaults to 0, is set to 1 after the hash-
				* table was filled with the information from
				* the pilot. */
  Tcl_HashTable db;  /* Refers to all known database objects, is indexed
		      * with the database names */

  /* -future-
   * conduit state information, ...
   */

} PiTclDlpInstance;


/*
 * This structure describes the per-instance state of a database object.
 * The system creates one instance per database on the pilot. Open databases
 * additionally have an object command accessing the structure.
 */

typedef struct PiTclDbInstance {

  PiTclDlpInstance* dlp; /* Backreference to the dlp object holding the
			  * connection to the pilot containing the database. */
  int         sd; /* Copy of the socket handle in the outer dlp object. A
		   * value of -1 indicates that the database is *not* open.
		   * In this case it must not try to access the socket during
		   * the close. */
  Tcl_Command cmd; /* Reference to our object command, for easy deletion.
		    * NULL for an unopened database. */
  int handle; /* The handle to the database, as returned by pilot-link upon
	       * opening it. -1 for a closed database. */
  DBInfo info; /* database information, as given to us by the pilot in the
		* cradle */
  int card;    /* database information missing in 'info', the card the
		* database resides on */

  Tcl_Obj* name;    /* redundant duplicate of the database name */
  Tcl_Obj* creator; /* redundant duplicate of the database creator */
  Tcl_Obj* type;    /* redundant duplicate of the database type */
  /* The duplicates above are used to speed up the search for a packer,
   * there is no need to create the objects on the fly */

  PiTclPackInfo* pack; /* Reference to the (un)packer commands to
			* apply to the data in this database. */

  /* Transport area */

  char* buf; /* Intermediate buffer for (un)packing data. 64 K in length,
	      * the max amount of data the pilot can handle. */
  int buflen; /* number of useful characters in the buffer */

  int isRecord; /* Integer flag, set if 'buf' contains record or resource
		 * information. (0 = app/sortinfo, 1 = record, 2 = resource) */
  int attr;     /* Record attributes, relevant only for record information */
  int category; /* Record category, relevant only for record information */
  recordid_t id; /* Record id, relevant only for record information */
  unsigned long resType; /* Resource type, relevant only for resource
			  * information. A 4 byte designator like database
			  * 'creator' and 'type'
			  */

} PiTclDbInstance;

#define MAXBUF (0xFFFF)







/* Utility functions dealing with DLP instance structures.
 */

PiTclDlpInstance*
PiTclCreateDlpObject _ANSI_ARGS_ ((PiTclInfo* info,
				   char*      deviceName));

int
PiTclOpenDlpChannel _ANSI_ARGS_ ((char* dev));

int
PiTclReleaseDlpObject _ANSI_ARGS_ ((PiTclDlpInstance* dlp));

int
PiTclDlpReadCards _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
				char*             cmd,
				Tcl_Interp*       interp));
int
PiTclDlpReadDbList _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
				 char*             cmd,
				 Tcl_Interp*       interp));

PiTclDbInstance*
PiTclCreateDbObject _ANSI_ARGS_ ((PiTclDlpInstance* dlp,
				  DBInfo*           dbInfo,
				  int               card));

void
PiTclReleaseDbObject _ANSI_ARGS_ ((PiTclDbInstance* db));

int
PiTclDlpDbCloseAll _ANSI_ARGS_ ((PiTclDlpInstance*   o,
				 char*             cmd,
				 Tcl_Interp*    interp,
				 int            mode));

enum PiTclCloseMode {
  PITCL_DB_CLOSE, PITCL_DB_DESTROY
};



/* Conversion of pilot datastructures from and to tcl lists.
 */

/* Card information
 */

Tcl_Obj*
PiTclCardToList _ANSI_ARGS_ ((Tcl_Interp* interp,
			      CardInfo*   card));
Tcl_Obj*
PiTclCardsToList _ANSI_ARGS_ ((Tcl_Interp*      interp,
			       PiTclDlpInstance* dlp));
/* Netsync information
 */

Tcl_Obj*
PiTclNetSyncInfoToList _ANSI_ARGS_ ((Tcl_Interp*  interp,
				     NetSyncInfo* nsi));

int
PiTclListToNetSyncInfo _ANSI_ARGS_ ((Tcl_Interp*  interp,
				     Tcl_Obj*     list,
				     NetSyncInfo* nsi));
/* System information
 */

Tcl_Obj*
PiTclSysInfoToList _ANSI_ARGS_ ((Tcl_Interp* interp,
				 SysInfo*    si));
/* PilotUser
 */

Tcl_Obj*
PiTclPilotUserToList _ANSI_ARGS_ ((Tcl_Interp* interp,
				    PilotUser*  pu));

int
PiTclListToPilotUser _ANSI_ARGS_ ((Tcl_Interp* interp,
				   Tcl_Obj*    list,
				   PilotUser*  pu));
/* -- -- */

int
PiTclListDatabases _ANSI_ARGS_ ((Tcl_Interp*      interp,
				 PiTclDlpInstance* dlp));

int
PiTclDbObjectToList _ANSI_ARGS_ ((Tcl_Interp*      interp,
				  PiTclDbInstance* db));




/* Conversion of string argument to long number (4 bytes)
 */

long int
PiTclStringToLong _ANSI_ARGS_ ((Tcl_Obj* o));

Tcl_Obj*
PiTclLongToString _ANSI_ARGS_ ((long int l));


/*
 * Procedures implementing the database object command.
 */

void
PiTclDeleteDbObjectCmd _ANSI_ARGS_ ((ClientData cd));

int
PiTclDlpDbObjectCmd _ANSI_ARGS_ ((ClientData     cd,
				  Tcl_Interp*    interp,
				  int            objc,
				  Tcl_Obj* CONST objv []));

/*
 * Procedures registering the wrappers to the existing packers.
 */

int
PiTclInitMemo _ANSI_ARGS_ ((Tcl_Interp* interp));

int
PiTclInitTodo _ANSI_ARGS_ ((Tcl_Interp* interp));

int
PiTclInitAppointment _ANSI_ARGS_ ((Tcl_Interp* interp));

int
PiTclInitAddress _ANSI_ARGS_ ((Tcl_Interp* interp));


/* Conversion helpers
 */

int
PiTclEncodeRecAttr _ANSI_ARGS_ ((Tcl_Interp* interp,
				 Tcl_Obj*    list,
				 int*        attr));

Tcl_Obj*
PiTclDecodeRecAttr _ANSI_ARGS_ ((Tcl_Interp* interp,
				 int         attr));

int
PiTclEncodeOpenFlags _ANSI_ARGS_ ((Tcl_Interp* interp,
				   Tcl_Obj*    list,
				   int*        flags));

int
PiTclEncodeDbFlags _ANSI_ARGS_ ((Tcl_Interp* interp,
				 Tcl_Obj*    list,
				 int*        flags));
Tcl_Obj*
PiTclDecodeDbFlags _ANSI_ARGS_ ((Tcl_Interp* interp,
				 int         flags));


#if 0
/* ******* Not yet, wrap the high-level interfaces first. */

/* C-level implementation of the 'piSocket' command.
 */

EXTERN int
PiTclChannelCmd _ANSI_ARGS_ ((ClientData notUsed,
			      Tcl_Interp* interp,
			      int argc, char** argv));
#endif /* 0 */

#ifdef __cplusplus
}
#endif /* C++ */
#endif /* PITCL_INT_H */
