/*
 * map.c --
 *
 *	Implementation of packer mapping management
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: map.c,v 1.1 1999/03/27 13:26:33 aku Exp $
 */


#include "map.h"
#include "macros.h"


static void
PiTclMapEntryFree _ANSI_ARGS_ ((PiTclMapEntry* e));

static void
PiTclMapDeleteList _ANSI_ARGS_ ((PiTclMapNode* n));

static int
CountStars _ANSI_ARGS_ ((PiTclMapEntry* e));

static int
AddEntry _ANSI_ARGS_ ((PiTclMapNode** list,
		       PiTclMapEntry* e));

static int
RemoveEntry _ANSI_ARGS_ ((Tcl_Interp*    interp,
			  PiTclMapNode** list,
			  PiTclMapEntry* e));

static int
EntryCompare _ANSI_ARGS_ ((PiTclMapEntry* a,
			   PiTclMapEntry* b));

static int
Starry _ANSI_ARGS_ ((Tcl_Obj* pattern));

static int
CompareStringObj _ANSI_ARGS_ ((Tcl_Obj* a,
			       Tcl_Obj* b));

static int
CompareStringObjStar _ANSI_ARGS_ ((Tcl_Obj* a,
				   Tcl_Obj* b));

static PiTclMapEntry*
FindEntry _ANSI_ARGS_ ((PiTclMapNode* n,
			Tcl_Obj*      db,
			Tcl_Obj*      creator,
			Tcl_Obj*      type));

static int
EntryMatch _ANSI_ARGS_ ((PiTclMapEntry* e,
			Tcl_Obj*        db,
			Tcl_Obj*        creator,
			Tcl_Obj*        type));

/*
 *------------------------------------------------------*
 *
 *      PiTclMapInit --
 *
 *	------------------------------------------------*
 *	This procedure initializes the mapping from
 *	databases to converters.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Modifies the given structure.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclMapInit (m)
PiTclMap* m;
{
  int i;

  for (i=0;i<4;i++) {
    m->list [i]  = (PiTclMapNode*) NULL;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclMapDelete --
 *
 *	------------------------------------------------*
 *	This procedure frees all resources of the given
 *	database -> converter map.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases memory.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclMapDelete (m)
PiTclMap* m;
{
  int i;

  for (i=0;i<4;i++) {
    PiTclMapDeleteList (m->list [i]);
    m->list [i]  = (PiTclMapNode*) NULL;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclMapDeleteList --
 *
 *	------------------------------------------------*
 *	Helper procedure for 'PiTclMapDelete'. Releases
 *	a single node list of the map.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases memory.
 *
 *	Result:
 *		Nothing.
 *
 *------------------------------------------------------*
 */

static void
PiTclMapDeleteList (n)
PiTclMapNode* n;
{
  PiTclMapNode* tmp;

  while (n != (PiTclMapNode*) NULL) {
    PiTclMapEntryFree (n->entry);
    
    tmp = n->next;
    Tcl_Free ((char*) n);
  }
}

/*
 *------------------------------------------------------*
 *
 *	PiTclMapEntryFree --
 *
 *	------------------------------------------------*
 *	Helper procedure, releases the information in a
 *	single entry of the map.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases memory.
 *
 *	Result:
 *		Nothing.
 *
 *------------------------------------------------------*
 */

static void
PiTclMapEntryFree (e)
PiTclMapEntry* e;
{
  Tcl_DecrRefCount (e->database);
  Tcl_DecrRefCount (e->creator);
  Tcl_DecrRefCount (e->type);
  PiTclPackRelease (e->pack);

  Tcl_Free ((char*) e);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclMapAdd --
 *
 *	------------------------------------------------*
 *	Adds an entry to the given map. It is assumed
 *	that the given entry resides in volatile memory
 *	and is therefore copied.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Modifies the map, allocates memory.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclMapAdd (m, e)
PiTclMap*      m;
PiTclMapEntry* e;
{
  return AddEntry (&(m->list [CountStars (e)]), e);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclMapRemove --
 *
 *	------------------------------------------------*
 *	Searches for the given entry in the specified map
 *	and removes it from it.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Modifies the map, releases memory.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclMapRemove (interp, m, db, creator, type)
Tcl_Interp*    interp;
PiTclMap*      m;
Tcl_Obj*      db;
Tcl_Obj* creator;
Tcl_Obj*    type;
{
  PiTclMapEntry e;

  e.database = db;
  e.creator  = creator;
  e.type     = type;

  return RemoveEntry (interp, &(m->list [CountStars (&e)]), &e);
}

/*
 *------------------------------------------------------*
 *
 *	PiTclMapFind --
 *
 *	------------------------------------------------*
 *	Searches the map for the entry which matches best
 *	database name, creator and type.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Nothing.
 *
 *	Result:
 *		A reference to the best map entry,
 *		or NULL
 *
 *------------------------------------------------------*
 */

PiTclMapEntry*
PiTclMapFind (m, db, creator, type)
PiTclMap*      m;
Tcl_Obj*      db;
Tcl_Obj* creator;
Tcl_Obj*    type;
{
  PiTclMapEntry* e;
  int            i;

  for (i=0;i<4;i++) {
    e = FindEntry (m->list [i], db, creator, type);
    if (e != (PiTclMapEntry*) NULL) {
      return e;
    }
  }

  return NULL;
}

/*
 *------------------------------------------------------*
 *
 *	AddEntry --
 *
 *	------------------------------------------------*
 *	This procedure adds the given entry to the list.
 *	An possibly existing entry is overwritten.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory, changes the structure of
 *		the list.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
AddEntry (list, e)
PiTclMapNode** list;
PiTclMapEntry* e;
{
  int new = 0;
  int delta = 0;

  if ((*list) == (PiTclMapNode*) NULL) {
    new = 1;
  }

  if (!new) {
    delta = EntryCompare (e, (*list)->entry);
  }

  if (new || (delta < 0)) {

    /* The new entry must be placed before the current one.
     * Allocate and initialize a copy of the transient entry.
     */

    PiTclMapEntry* ne = (PiTclMapEntry*) Tcl_Alloc (sizeof (PiTclMapEntry));
    PiTclMapNode*  n  = (PiTclMapNode*)  Tcl_Alloc (sizeof (PiTclMapNode));

    ne->database = e->database;
    ne->creator  = e->creator;
    ne->type     = e->type;
    ne->pack     = e->pack;

    Tcl_IncrRefCount  (ne->database);
    Tcl_IncrRefCount  (ne->creator);
    Tcl_IncrRefCount  (ne->type);
    PiTclPackPreserve (ne->pack);

    n->entry = ne;
    n->next  = *list;
    (*list)  = n;

    return TCL_OK;

  } else if (delta == 0) {
    /* Overwrite converter in existing entry.
     */

    PiTclPackPreserve (e->pack);
    PiTclPackRelease  ((*list)->entry->pack);

    (*list)->entry->pack = e->pack;

    return TCL_OK;
  } else {
    return AddEntry (&((*list)->next), e);
  }
}

/*
 *------------------------------------------------------*
 *
 *	RemoveEntry --
 *
 *	------------------------------------------------*
 *	This procedure searches the given entry in the
 *	list and removes it from it.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Releases memory, changes the structure of
 *		the list.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

static int
RemoveEntry (interp, list, e)
Tcl_Interp*    interp;
PiTclMapNode** list;
PiTclMapEntry* e;
{
  int delta;

  if ((*list) == (PiTclMapNode*) NULL) {
    Tcl_AppendResult (interp, "entry not found", (char*) NULL);
    return TCL_ERROR;
  }

  delta = EntryCompare (e, (*list)->entry);

  if (delta > 0) {
    /* Entry is greater than current in list, we are behind the
     * possible insertion point, not found, signal error.
     */

    Tcl_AppendResult (interp, "entry not found", (char*) NULL);
    return TCL_ERROR;
  } else if (delta == 0) {
    /* Found it, splice out of the list */

    PiTclMapNode* n  = (*list);

    (*list) = n->next;

    PiTclMapEntryFree (n->entry);
    Tcl_Free ((char*) n);

    return TCL_OK;
  } else {
    /* Search on */

    return RemoveEntry (interp, &((*list)->next), e);
  }
}

/*
 *------------------------------------------------------*
 *
 *      CountStars --
 *
 *	------------------------------------------------*
 *	This procedure counts the number of variable
 *	patterns in the given map entry.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None
 *
 *	Result:
 *		An integer number in the range 0 .. 3.
 *
 *------------------------------------------------------*
 */

static int
CountStars (e)
PiTclMapEntry* e;
{
  return (Starry (e->database) +
	  Starry (e->creator)  +
	  Starry (e->type));
}

/*
 *------------------------------------------------------*
 *
 *	EntryCompare --
 *
 *	------------------------------------------------*
 *	This procedure compares to map entries.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		-1, 0, 1, depending on wether a is less
 *		than, equal to or greather than b.
 *
 *------------------------------------------------------*
 */

static int
EntryCompare (a, b)
PiTclMapEntry* a;
PiTclMapEntry* b;
{
  int delta;

  delta = CompareStringObjStar (a->database, b->database);

  if (delta != 0) {
    return delta;
  }

  delta = CompareStringObjStar (a->creator, b->creator);

  if (delta != 0) {
    return delta;
  }

  return CompareStringObjStar (a->type, b->type);
}

/*
 *------------------------------------------------------*
 *
 *	Starry --
 *
 *	------------------------------------------------*
 *	This procedure determines wether a string contains
 *	the 'asterisk' character (*, star) or not.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		Boolean true / false encoded as 1 and 0.
 *
 *------------------------------------------------------*
 */

static int
Starry (pattern)
Tcl_Obj* pattern;
{
  int   len, i;
  char* str = Tcl_GetStringFromObj (pattern, &len);

  for (i=0;i<len;i++) {
    if (str [i] == '*') {
      return 1;
    }
  }

  return 0;
}

/*
 *------------------------------------------------------*
 *
 *	CompareStringObjStar --
 *
 *	------------------------------------------------*
 *	This procedure compares the string representations
 *	of two tcl objects. Specialty: A string containing
 *	at least one star ranks higher than an unstarred
 *	string.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		-1, 0, 1, depending on wether a is less
 *		than, equal to or greather than b.
 *
 *------------------------------------------------------*
 */

static int
CompareStringObjStar (a, b)
Tcl_Obj* a;
Tcl_Obj* b;
{
  int astar = Starry (a);
  int bstar = Starry (b);

  if (astar == bstar) {
    /* both have stars, or not, lexical order */
    return CompareStringObj (a, b);
  }

  /* Only one has a star, that iis the greater one */

  if (astar) {
    return 1;
  } else {
    return -1;
  }  
}

/*
 *------------------------------------------------------*
 *
 *	CompareStringObj --
 *
 *	------------------------------------------------*
 *	This procedure compares the string representations
 *	of two tcl objects.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		-1, 0, 1, depending on wether a is less
 *		than, equal to or greather than b.
 *
 *------------------------------------------------------*
 */

static int
CompareStringObj (a, b)
Tcl_Obj* a;
Tcl_Obj* b;
{
  int lenA, lenB, length, match;

  char* strA = Tcl_GetStringFromObj (a, &lenA);
  char* strB = Tcl_GetStringFromObj (b, &lenB);

  length = (lenA < lenB) ? lenA : lenB;
  match  = memcmp (strA, strB, (unsigned) length);

  if (match == 0) {
    match = lenA - lenB;
  }

  return ((match > 0) ? 1 : ((match < 0) ? -1 : 0));
}

/*
 *------------------------------------------------------*
 *
 *	FindEntry --
 *
 *	------------------------------------------------*
 *	This procedure searches a matching entry in the
 *	given list, the first found is returned
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		A reference to the matching entry, or NULL.
 *
 *------------------------------------------------------*
 */

static PiTclMapEntry*
FindEntry (n, db, creator, type)
PiTclMapNode* n;
Tcl_Obj*      db;
Tcl_Obj*      creator;
Tcl_Obj*      type;
{
  if (n == (PiTclMapNode*) NULL) {
    return (PiTclMapEntry*) NULL;
  }

  if (EntryMatch (n->entry, db, creator, type)) {
    return n->entry;
  }

  return FindEntry (n->next, db, creator, type);
}

/*
 *------------------------------------------------------*
 *
 *	EntryMatch --
 *
 *	------------------------------------------------*
 *	This procedure checks wether the entry matches
 *	the database information.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		None.
 *
 *	Result:
 *		A boolean value, 0/FALSE, 1/TRUE.
 *
 *------------------------------------------------------*
 */

static int
EntryMatch (e, db, creator, type)
PiTclMapEntry* e;
Tcl_Obj*       db;
Tcl_Obj*       creator;
Tcl_Obj*       type;
{
  /* -W- Beware: The tcl procedure used here is *not* 8bit
   * -W- clean and will mistreat strings containing '\0' !
   * -W- This might be critical for 'creator' and 'type',
   * -W- which are allowed to have such.
   */

  if (!Tcl_StringMatch (Tcl_GetStringFromObj (db,          NULL),
			Tcl_GetStringFromObj (e->database, NULL))) {
    return 0;
  }

  if (!Tcl_StringMatch (Tcl_GetStringFromObj (creator,    NULL),
			Tcl_GetStringFromObj (e->creator, NULL))) {
    return 0;
  }

  if (!Tcl_StringMatch (Tcl_GetStringFromObj (type,    NULL),
			Tcl_GetStringFromObj (e->type, NULL))) {
    return 0;
  }

  return 1;
}
/*
 *------------------------------------------------------*
 *
 *	PiTclMapClear --
 *
 *	------------------------------------------------*
 *	Removes all entries from the mapping.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Modifies the map, releases memory.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclMapClear (interp, m)
Tcl_Interp*    interp;
PiTclMap*      m;
{
  int           i;
  PiTclMapNode* n;
  PiTclMapNode* nn;

  for (i=0;i<4;i++) {
    n = m->list [i];

    while (n != (PiTclMapNode*) NULL) {
      nn = n->next;

      PiTclMapEntryFree (n->entry);
      Tcl_Free ((char*) n);

      n = nn;
    }

    m->list [i] = (PiTclMapNode*) NULL;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	PiTclMapList --
 *
 *	------------------------------------------------*
 *	Lists all entries in the map.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Sets the result of the specified
 *		interpreter to the constructed list.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
PiTclMapList (interp, m)
Tcl_Interp*    interp;
PiTclMap*      m;
{
  int           i, res;
  PiTclMapNode* n;
  Tcl_Obj*      list = Tcl_NewListObj (0, NULL);
  Tcl_Obj*      elist;

  for (i=0;i<4;i++) {
    n = m->list [i];

    while (n != (PiTclMapNode*) NULL) {

      elist = Tcl_NewListObj (0, NULL);

      LIST_ADDOBJ (elist, n->entry->database);
      LIST_ADDOBJ (elist, n->entry->creator);
      LIST_ADDOBJ (elist, n->entry->type);
      LIST_ADDSTR (elist, n->entry->pack->name);

      LIST_ADDOBJ (list, elist);

      n = n->next;
    }
  }

  Tcl_SetObjResult (interp, list);
  return TCL_OK;
}

