/* -*- c -*-
 *
 * tclUtil.h --
 *
 *	General utilities helping in the construction of extensions
 *
 * Copyright (C) 1998 Andreas Kupries (a.kupries@westend.com)
 * All rights reserved.
 *
 * Permission is hereby granted, without written agreement and without
 * license or royalty fees, to use, copy, modify, and distribute this
 * software and its documentation for any purpose, provided that the
 * above copyright notice and the following two paragraphs appear in
 * all copies of this software.
 *
 * IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT, SPECIAL,
 * INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OF THIS
 * SOFTWARE AND ITS DOCUMENTATION, EVEN IF I HAVE BEEN ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * I SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND
 * I HAVE NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
 * ENHANCEMENTS, OR MODIFICATIONS.
 *
 * CVS: $Id: tclUtil.c,v 1.1 1999/03/27 13:26:34 aku Exp $
 */

#include "tclUtil.h"

/*
 *------------------------------------------------------*
 * Forward declarations of internal procedures.
 */

static void
CompleteBitsetDef _ANSI_ARGS_ ((TclUtil_Bitset* setDef));

static void
CompleteEnumDef _ANSI_ARGS_ ((TclUtil_Enum* enumDef));

static void
CompleteStructDef _ANSI_ARGS_ ((TclUtil_Struct* structDef));

static Tcl_Obj*
StructDecodeItem _ANSI_ARGS_ ((Tcl_Interp*         interp,
			       TclUtil_StructItem* structItem,
			       void *              structPtr));

static int
StructEncodeItem _ANSI_ARGS_ ((Tcl_Interp*         interp,
			       TclUtil_StructItem* structItem,
			       void *              structPtr,
			       Tcl_Obj*            item));

/*
 *------------------------------------------------------*
 *
 *	TclUtil_BitsetEncode --
 *
 *	------------------------------------------------*
 *	This procedure converts a list containing the
 *	names of bits in a set into the actual bitset,
 *	an integer number.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Changes the contents of argument 'set'.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
TclUtil_BitsetEncode (interp, setDef, setList, set)
Tcl_Interp*     interp;  /* Interpreter to use for error messages */
TclUtil_Bitset* setDef;  /* Definition of the bitset */
Tcl_Obj*        setList; /* List containing the names of the set bits */
int*            set;     /* Place to store the result of the encoding */
{
  int       result = 0;
  Tcl_Obj** lobjs;
  int       length, index;

  CompleteBitsetDef (setDef);

  if (Tcl_ListObjGetElements (interp, setList,
			      &length, &lobjs) != TCL_OK) {
    return TCL_ERROR;
  }

  for (length --; length >= 0; length --) {
    if (Tcl_GetIndexFromObj (interp, lobjs [length],
			     setDef->texts, "bit", 0,
			     &index) != TCL_OK) {
      return TCL_ERROR;
    }
    
    result |= setDef->bits [index].bit;
  }

  /*  printf ("Encode (%s, %s) = %d\n", Tcl_GetStringFromObj (setList, NULL), setDef->name, result); */

  *set = result;
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	TclUtil_BitsetDecode --
 *
 *	------------------------------------------------*
 *	This procedure converts a bitset into a list
 *	containing the names of the set bits.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates tcl objects.
 *
 *	Result:
 *		A reference to the created list object,
 *		or NULL.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
TclUtil_BitsetDecode (interp, setDef, set)
Tcl_Interp*     interp;  /* Interpreter to use for error messages */
TclUtil_Bitset* setDef;  /* Definition of the bitset to convert */
int             set;     /* The bitset to convert */
{
  Tcl_Obj* str;
  Tcl_Obj* list = Tcl_NewListObj (0, NULL);
  int      res, i;

  CompleteBitsetDef (setDef);

  for (i=0; setDef->texts [i] != (char*) NULL; i++) {
    if (set & setDef->bits [i].bit) {

      str = Tcl_NewStringObj (setDef->texts [i], -1);

      if (str == (Tcl_Obj*) NULL) {
	Tcl_DecrRefCount (list);
	return (Tcl_Obj*) NULL;
      }

      res = Tcl_ListObjAppendElement (interp, list, str);

      if (res != TCL_OK) {
	Tcl_DecrRefCount (list);
	return (Tcl_Obj*) NULL;
      }
    }
  }

  /* printf ("Decode (%d, %s) = %s\n", set, setDef->name,
     Tcl_GetStringFromObj (list, NULL)); */

  return list;
}

/*
 *------------------------------------------------------*
 *
 *	CompleteBitsetDef --
 *
 *	------------------------------------------------*
 *	This procedure initializes the 'texts' array of
 *	a bitset definition, if that is necessary.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory, which is never released!
 *
 *	Result:
 *		Nothing
 *
 *------------------------------------------------------*
 */

static void
CompleteBitsetDef (setDef)
TclUtil_Bitset* setDef;
{
  int i, len;

  if (setDef->texts != (char**) NULL) {
    return;
  }

  /* Count the number of bits in the set, then allocate
   * the 'texts'-array with the appropriate size, at last
   * copy the references.
   */

  for (i=0, len=1; setDef->bits [i].text != (char*) NULL; i++) {
    len ++;
  }

  setDef->texts = (char**) Tcl_Alloc (len * sizeof (char*));

  for (i=0; i < len; i++) {
    setDef->texts [i] = setDef->bits [i].text;
  }
}

/*
 *------------------------------------------------------*
 *
 *	TclUtil_EnumEncode --
 *
 *	------------------------------------------------*
 *	This procedure converts a string containing the
 *	names of a value in an enum into the actual enum
 *	value, an integer number.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Changes the contents of argument 'enValue'.
 *
 *	Result:
 *		A standard Tcl result.
 *
 *------------------------------------------------------*
 */

int
TclUtil_EnumEncode (interp, enumDef, o, enValue)
Tcl_Interp*   interp;  /* Interpreter to use for error messages */
TclUtil_Enum* enumDef; /* Definition of the enum */
Tcl_Obj*      o;       /* String containing the name of an enum value */
int*          enValue; /* Place to store the result of the encoding */
{
  int index;

  CompleteEnumDef (enumDef);

  if (Tcl_GetIndexFromObj (interp, o,
			   enumDef->texts,
			   enumDef->name, 0,
			   &index) != TCL_OK) {
      return TCL_ERROR;
  }

  /*  printf ("Encode (%s, %s) = %d\n", Tcl_GetStringFromObj (o, NULL), enumDef->name, index); */

  *enValue = enumDef->value [index].value;
  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	TclUtil_EnumDecode --
 *
 *	------------------------------------------------*
 *	This procedure converts an enum value into a string
 *	containing the name of the value.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates tcl objects.
 *
 *	Result:
 *		A reference to the created string object,
 *		or NULL.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
TclUtil_EnumDecode (interp, enumDef, enValue)
Tcl_Interp*   interp;  /* Interpreter to use for error messages */
TclUtil_Enum* enumDef; /* Definition of the enum to use for conversion */
int           enValue; /* The enum value to convert */
{
  Tcl_Obj* str;
  int      i;

  CompleteEnumDef (enumDef);

  for (i=0; enumDef->texts [i] != (char*) NULL; i++) {
    if (enumDef->value [i].value == enValue) {
      str = Tcl_NewStringObj (enumDef->value [i].text, -1);

      if (str == (Tcl_Obj*) NULL) {
	return (Tcl_Obj*) NULL;
      }

      /* printf ("Decode (%d, %s) = %s\n", enValue, enumDef->name,
	 Tcl_GetStringFromObj (str, NULL)); */

      return str;
    }
  }

  /* Not found */
  Tcl_AppendResult (interp, "unknown enumeration value", NULL);
  return (Tcl_Obj*) NULL;
}

/*
 *------------------------------------------------------*
 *
 *	CompleteEnumDef --
 *
 *	------------------------------------------------*
 *	This procedure initializes the 'texts' array of
 *	an enum definition, if that is necessary.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory, which is never released!
 *
 *	Result:
 *		Nothing
 *
 *------------------------------------------------------*
 */

static void
CompleteEnumDef (enumDef)
TclUtil_Enum* enumDef;
{
  int i, len;

  if (enumDef->texts != (char**) NULL) {
    return;
  }

  /* Count the number of values in the enumeration, then
   * allocate the 'texts'-array with the appropriate size,
   * at last copy the references.
   */

  for (i=0, len=1; enumDef->value [i].text != (char*) NULL; i++) {
    len ++;
  }

  enumDef->texts = (char**) Tcl_Alloc (len * sizeof (char*));

  for (i=0; i < len; i++) {
    enumDef->texts [i] = enumDef->value [i].text;
  }
}

/*
 *------------------------------------------------------*
 *
 *	CompleteStructDef --
 *
 *	------------------------------------------------*
 *	This procedure initializes the 'texts' array of
 *	a structure definition, if that is necessary.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Allocates memory which is never released!
 *
 *	Result:
 *		Nothing
 *
 *------------------------------------------------------*
 */

static void
CompleteStructDef (structDef)
TclUtil_Struct* structDef;
{
  int i, len;

  if (structDef->texts != (char**) NULL) {
    return;
  }

  /* Count the number of items in the structure, then allocate
   * the 'texts'-array with the appropriate size, at last
   * copy the references.
   */

  for (i=0, len=1; structDef->items [i].name != (char*) NULL; i++) {
    len ++;
  }

  structDef->texts = (char**) Tcl_Alloc (len * sizeof (char*));
  structDef->allItemMask = 0;

  for (i=0; i < len; i++) {
    structDef->texts [i] = structDef->items [i].name;
    structDef->allItemMask |= BIT (i);
  }

  /* Don't count ITEM_END */
  structDef->allItemMask &= ~(BIT (len-1));
}



int
TclUtil_StructEncode (interp, structDef, structList, structPtr)
Tcl_Interp*     interp;
TclUtil_Struct* structDef;
Tcl_Obj*        structList;
void*           structPtr;
{
  Tcl_Obj** lobjs;
  int       length, key;
  long int  items = 0;

  CompleteStructDef (structDef);

  if (Tcl_ListObjGetElements (interp, structList,
			      &length, &lobjs) != TCL_OK) {
    return TCL_ERROR;
  }

  if (length % 2 == 1) {
    Tcl_AppendResult (interp,
		      "length of list is no multiple of 2",
		      (char*) NULL);
    return TCL_ERROR;
  }

  while (length > 0) {
    if (Tcl_GetIndexFromObj(interp, lobjs [0],
			    structDef->texts,
			    structDef->name,
			    0, &key) != TCL_OK) {
      return TCL_ERROR;
    }

    if (StructEncodeItem (interp, &structDef->items [key],
			  structPtr, lobjs [1]) != TCL_OK) {
      return TCL_ERROR;
    }

    /* Step to next pair, include encountered part in enum */

    length -= 2;
    lobjs  += 2;
    items  |= BIT (key);
  }

  if (items != structDef->allItemMask) {
    Tcl_AppendResult (interp,
		      "some keys missing in list",
		      (char*) NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}

/*
 *------------------------------------------------------*
 *
 *	TclUtil_StructDecode --
 *
 *	------------------------------------------------*
 *	This procedure converts a structure into a
 *	readable Tcl_Obj (list), suitable for 'array set'.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates Tcl_Obj's.
 *
 *	Result:
 *		A reference to the created Tcl_Obj,
 *		or NULL.
 *
 *------------------------------------------------------*
 */

Tcl_Obj*
TclUtil_StructDecode (interp, structDef, structPtr)
Tcl_Interp*     interp;
TclUtil_Struct* structDef;
void*           structPtr;
{
  Tcl_Obj* result = Tcl_NewListObj (0, NULL);
  Tcl_Obj* item;
  int      i, res;

  CompleteStructDef (structDef);

  for (i=0; structDef->items [i].name != (char*) NULL; i++) {

    item = Tcl_NewStringObj (structDef->items [i].name, -1);

    if (item == (Tcl_Obj*) NULL) {
      Tcl_AppendResult (interp,
			"no memory for item ",
			structDef->items [i].name,
			NULL);
      Tcl_DecrRefCount (result);
      return (Tcl_Obj*) NULL;
    }

    res = Tcl_ListObjAppendElement (interp, result, item);

    if (res != TCL_OK) {
      Tcl_AppendResult (interp,
			"could not append name of item ",
			structDef->items [i].name,
			NULL);
      Tcl_DecrRefCount (result);
      return (Tcl_Obj*) NULL;
    }

    item = StructDecodeItem (interp, &structDef->items [i],
			     structPtr);

    if (item == (Tcl_Obj*) NULL) {
      Tcl_AppendResult (interp,
			"could not convert item ",
			structDef->items [i].name,
			NULL);
      Tcl_DecrRefCount (result);
      return (Tcl_Obj*) NULL;
    }

    res = Tcl_ListObjAppendElement (interp, result, item);

    if (res != TCL_OK) {
      Tcl_AppendResult (interp,
			"could not append value of item ",
			structDef->items [i].name,
			NULL);
      Tcl_DecrRefCount (result);
      return (Tcl_Obj*) NULL;
    }
  }

  return result;
}

/*
 *------------------------------------------------------*
 *
 *	StructDecodeItem --
 *
 *	------------------------------------------------*
 *	This procedure converts an item in a structure
 *	into a readable Tcl_Obj.
 *	------------------------------------------------*
 *
 *	Sideeffects:
 *		Creates an Tcl_Obj.
 *
 *	Result:
 *		A reference to the created Tcl_Obj,
 *		or NULL.
 *
 *------------------------------------------------------*
 */

static Tcl_Obj*
StructDecodeItem (interp, structItem, structPtr)
Tcl_Interp*         interp;
TclUtil_StructItem* structItem;
void *              structPtr;
{
  char* itemPtr =         ((char*) structPtr) + structItem->offset;
  int*  lenPtr  = (int*) (((char*) structPtr) + structItem->lenOffset);

  switch (structItem->type) {
  case TUSI_Int:     /* A standard integer */
    return Tcl_NewIntObj (*((int*) itemPtr));

  case TUSI_Byte:     /* An 8-bit integer */
    return Tcl_NewIntObj ((int) *((char*) itemPtr));

  case TUSI_Boolean: /* integer variant, contains a boolean value */
    return Tcl_NewIntObj (((*((int*) itemPtr)) != 0) ? 1 : 0);

  case TUSI_Double:  /* A standard double */
    return Tcl_NewDoubleObj (*((double*) itemPtr));

  case TUSI_Chars:    /* Inlined string, \0-terminated, at most 'n' chars */
    return Tcl_NewStringObj (itemPtr, -1);

  case TUSI_PChars:    /* Inlined string, length in separate field */
    return Tcl_NewStringObj (itemPtr, *lenPtr);

  case TUSI_CharN:   /* exactly 'n' characters, inlined */
    return Tcl_NewStringObj (itemPtr, structItem->length);

  case TUSI_String:  /* A pointer to a character string */
    return Tcl_NewStringObj (*((char**) itemPtr), -1);

  case TUSI_Bitset:  /* A TclUtil_Bitset governs conversion, integer variant */
    return TclUtil_BitsetDecode (interp,
				 structItem->clientData,
				 *((int*) itemPtr));

  case TUSI_Struct:  /* A TclUtil_Struct governs conversion of sub structure */
    return TclUtil_StructDecode (interp,
				 structItem->clientData,
				 itemPtr);

  case TUSI_Enum:  /* A TclUtil_Enum governs conversion of integer */
    return TclUtil_EnumDecode (interp,
			       structItem->clientData,
			       *((int*) itemPtr));

  case TUSI_Custom:   /* Type for items currently not known,
		       * custom converter procedures are required. */
    return (structItem->toObjProc) (interp,
				    (void*) itemPtr,
				    structItem->clientData);

  default:
    return (Tcl_Obj*) NULL;
  }
}



static int
StructEncodeItem (interp, structItem, structPtr, item)
Tcl_Interp*         interp;
TclUtil_StructItem* structItem;
void *              structPtr;
Tcl_Obj*            item;
{
  char* itemPtr =         ((char*) structPtr) + structItem->offset;
  int*  lenPtr  = (int*) (((char*) structPtr) + structItem->lenOffset);

  switch (structItem->type) {
  case TUSI_Int:     /* A standard integer */
    if (Tcl_GetIntFromObj (interp, item, (int*) itemPtr) != TCL_OK) {
      return TCL_ERROR;
    }
    break;

  case TUSI_Byte:     /* An 8-bit integer */
    {
      int value;

      if (Tcl_GetIntFromObj (interp, item, &value) != TCL_OK) {
	return TCL_ERROR;
      }

      *((unsigned char*) itemPtr) = (unsigned char) value;
    }
    break;

  case TUSI_Boolean: /* integer variant, contains a boolean value */
    {
      int v;
      if (Tcl_GetIntFromObj (interp, item, &v) != TCL_OK) {
	return TCL_ERROR;
      }
      * ((int*) itemPtr) = (v != 0) ? 1 : 0;
    }
    break;

  case TUSI_Double:  /* A standard double */
    if (Tcl_GetDoubleFromObj (interp, item, (double*) itemPtr) != TCL_OK) {
      return TCL_ERROR;
    }
    break;

  case TUSI_PChars: /* Inlined string, at most 'n' chars, len in sep. field */
    {
      int   len;
      char* str = Tcl_GetStringFromObj (item, &len);
      
      if (len > structItem->length) {
	len = structItem->length;
      }

      memcpy (itemPtr, str, len);
      *lenPtr = len;
    }
  break;

  case TUSI_Chars:    /* Inlined string, at most 'n' chars, \0-term. */
  case TUSI_CharN:   /* Exactly 'n' characters, inlined */
    {
      int   len;
      char* str = Tcl_GetStringFromObj (item, &len);
      
      if (len > (structItem->length-1)) {
	len = structItem->length-1;
      }

      strncpy (itemPtr, str, len);
      itemPtr [len] = '\0';
    }
  break;

  case TUSI_Bitset:  /* A TclUtil_Bitset governs conversion, integer variant */
    return TclUtil_BitsetEncode (interp,
				 structItem->clientData,
				 item,
				 (int*) itemPtr);
  
  case TUSI_Struct:  /* A TclUtil_Struct governs conversion of sub structure */
    return TclUtil_StructEncode (interp,
				 structItem->clientData,
				 item,
				 itemPtr);
  
  case TUSI_Enum:  /* A TclUtil_Enum governs conversion of integer */
    return TclUtil_EnumEncode (interp,
			       structItem->clientData,
			       item,
			       (int*) itemPtr);
  
  case TUSI_String:  /* A pointer to a character string */
    /* Assume dynamic allocation of structure item */
    {
      char*  str    = Tcl_GetStringFromObj (item, NULL);
      char** strPtr = (char**) itemPtr;

      if (*strPtr != (char*) NULL) {
	Tcl_Free (*strPtr);
      }

      *strPtr = DUP (str);
    }
  break;

  case TUSI_Custom:   /* Type for items currently not known,
		       * custom converter procedures are required. */
    return (structItem->fromObjProc) (interp, item,
				      (void*) itemPtr,
				      structItem->clientData);

  default:
    return TCL_ERROR;
  }

  return TCL_OK;
}
